.sourceview,
.history-entry {
  border-top-color: @borders;
}

.math-popover list > row:hover,
.math-popover list > row:active {
  box-shadow: inset 0 0 0 1px @borders;
}

.math-popover .constant-list image {
  opacity: 80%;
}
